/* sqUnixSound.c -- sound support for various Unix sound systems
 *
 *   Copyright (C) 1996 1997 1998 1999 2000 2001 Ian Piumarta and individual
 *      authors/contributors listed elsewhere in this file.
 *   All rights reserved.
 *   
 *   This file is part of Unix Squeak.
 * 
 *   This file is distributed in the hope that it will be useful, but WITHOUT
 *   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *   FITNESS FOR A PARTICULAR PURPOSE.
 *   
 *   You may use and/or distribute this file ONLY as part of Squeak, under
 *   the terms of the Squeak License as described in `LICENSE' in the base of
 *   this distribution, subject to the following restrictions:
 * 
 *   1. The origin of this software must not be misrepresented; you must not
 *      claim that you wrote the original software.  If you use this software
 *      in a product, an acknowledgment to the original author(s) (and any
 *      other contributors mentioned herein) in the product documentation
 *      would be appreciated but is not required.
 * 
 *   2. This notice may not be removed or altered in any source distribution.
 * 
 *   Using or modifying this file for use in any context other than Squeak
 *   changes these copyright conditions.  Read the file `COPYING' in the base
 *   of the distribution before proceeding with any such use.
 * 
 *   You are STRONGLY DISCOURAGED from distributing a modified version of
 *   this file under its original name without permission.  If you must
 *   change it, rename it first.
 *
 * Authors: Ian.Piumarta@inria.fr and Lex Spoon <lex@cc.gatech.edu>
 */

#include "sq.h"
#include "aio.h"

int snd_Stop(void);

#if defined(USE_AUDIO_NAS)

/* NAS is the Network Audio System.  NAS uses a network sound server
 * whose streaming interface meshes very nicely with what Squeak needs.
 */

#include <audio/audiolib.h>

#define FAIL(X) { success(false); return X; }

static AuServer *server= NULL;   /* the audio server to write to */
static AuFlowID flow;
static int semaIndex;          /* the semaphore to signal squeak with */
static int stereo;             /* whether we are playing in stereo or not */
static int bytesAvail;         /* current number of bytes that may be written
				  to the server */

#if 0
# define DEBUG
#endif
#ifdef DEBUG
# define dprintf printf
#else
  static void dprintf(const char *fmt, ...) {}
#endif

static int bytesPerSample(void) 
{
  return stereo ? 4 : 2;
}


/*** sound output ***/


int snd_AvailableSpace(void)
{
  if (server == NULL)
    return 0;
  return bytesAvail;
}

     
int snd_InsertSamplesFromLeadTime(int frameCount, int srcBufPtr,
				  int samplesOfLeadTime)
{
  /* not possible, I don't think, across the network */
  success(false);
  return 0;
}


int snd_PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
{
  int bytesRequested;
  int bytesToPlay;
  int framesToPlay;
  
  bytesRequested= frameCount * bytesPerSample();
  bytesToPlay= bytesRequested;
  if (bytesToPlay > bytesAvail)
    bytesToPlay= bytesAvail;
  
  framesToPlay= bytesToPlay / bytesPerSample();

  dprintf("writing %d bytes (%d frames)\n", bytesToPlay, framesToPlay);
  AuWriteElement(server, flow, 0,
		 bytesToPlay,
		 (void *)(arrayIndex + (4 * startIndex)),
		 AuFalse,
		 NULL);
  AuFlush(server);

  bytesAvail-= bytesToPlay;
  
  return framesToPlay;
}


static void handleAudioEvents(int fd, void *data, int flags)
{
  if (!server)
    {
      dprintf( "handleAudioEvents called while unconnected!\n");
      return;
    }

  /* read events once */
  AuEventsQueued(server, AuEventsQueuedAfterReading);

  /* then loop through the read queue */
  while (AuEventsQueued(server, AuEventsQueuedAlready))
    {
      AuEvent event;
      AuNextEvent(server, AuTrue, &event);
      dprintf("event of type %d\n", event.type);
    
      switch(event.type)
	{
	case 0:
	  {
	    AuErrorEvent *errEvent= (AuErrorEvent *) &event;
	    char errdesc[1000];
      
	    AuGetErrorText(server, errEvent->error_code, errdesc, sizeof(errdesc));
	    fprintf(stderr, "audio error: %s\n", errdesc);
	    snd_Stop();
	    return;  /* return, not break, so that we don't
			process the now-closed server any longer! */
	  }
	case AuEventTypeElementNotify:
	  {
	    AuElementNotifyEvent *enEvent= (AuElementNotifyEvent *)&event;
	    switch(enEvent->kind)
	      {
	      case AuElementNotifyKindLowWater:
		dprintf("low water event\n");
		bytesAvail+= enEvent->num_bytes;
		break;
	      case AuElementNotifyKindHighWater:
		dprintf("high water event\n");
		bytesAvail+= enEvent->num_bytes;
		break;
	      case AuElementNotifyKindState:
		dprintf("state change (%d->%d)\n",
			enEvent->prev_state,
			enEvent->cur_state);
		bytesAvail+= enEvent->num_bytes;
		break;
	      }
	  }
	}
    }

  if (bytesAvail > 0)
    {
      dprintf("bytesAvail: %d\n", bytesAvail);
      signalSemaphoreWithIndex(semaIndex);
    }

  aioHandle(AuServerConnectionNumber(server), handleAudioEvents, AIO_R);
}


int snd_PlaySilence(int time)
{
  return 0;
}


static unsigned chooseNasDevice(AuServer *server, int samplesPerSec, int stereo)
{
  int desiredChannels= stereo ? 2 : 1;
  int i;
  
  /* look for a physical, output device with the proper number of channels */
  for (i= 0;  i < AuServerNumDevices(server);  i++)
    {
      if ((AuDeviceKind(AuServerDevice(server, i)) ==  AuComponentKindPhysicalOutput)
	  && (AuDeviceNumTracks(AuServerDevice(server, i)) ==  desiredChannels))
	return AuDeviceIdentifier(AuServerDevice(server, i));
    }
  return 0;   /* XXX hopefully this is an illegal device identifier */
}


int snd_Start(int frameCount, int samplesPerSec, int stereo0, int semaIndex0)
{
  AuElement elements[2];  /* first is a client element, second is
			     a device output element */
  unsigned device;        /* ID of the device to play to */
  
  /* open the server */
  dprintf("opening server\n");
  server= AuOpenServer(NULL, 0, NULL, 0, NULL, NULL);
  if (server == NULL)
    {
      dprintf("failed to open audio server\n");
      return false;
    }
  /* record requested info */
  semaIndex= semaIndex0;
  stereo= stereo0;
  /* pick a device to play to */ 
  device= chooseNasDevice(server, samplesPerSec, stereo);
  if (device == 0)
    {
      dprintf("no available device on the server!\n");
      AuCloseServer(server);
      server= NULL;
      return false;
    }
  /* create a flow to write on */
  dprintf("creating flow\n");
  flow= AuCreateFlow(server, NULL);
  /* create client and device elements to play with */
  dprintf("creating elements(%d,%d)\n", frameCount, frameCount / 4);
  AuMakeElementImportClient(&elements[0],
			    samplesPerSec,
			    AuFormatLinearSigned16LSB,  /* XXX this should be chosen based on the platform */
			    stereo ? 2 : 1,
			    AuTrue,
			    2*frameCount,   /* max */
			    frameCount,   /* low */
			    0, NULL);
  AuMakeElementExportDevice(&elements[1],
			    0,
			    device,
			    samplesPerSec,
			    AuUnlimitedSamples,
			    0, NULL);
  /* set up the flow with these elements */
  AuSetElements(server,	flow,
		AuTrue,
		2, elements,
		NULL);
  /* start her up */
  dprintf("starting flow\n");
  AuStartFlow(server, flow, NULL);
  AuFlush(server);
  /* initialize the space indication */
  bytesAvail= 0;
  /* arrange to be informed when events come in from the server */
  aioEnable(AuServerConnectionNumber(server), 0, 0);
  aioHandle(AuServerConnectionNumber(server), handleAudioEvents, AIO_R);

  return true;
}


int snd_Stop(void)
{
  if (server != NULL)
    {
      aioDisable(AuServerConnectionNumber(server));
      AuCloseServer(server);
      server= NULL;
    }
  return 0;
}


/*** sound input ***/


int snd_SetRecordLevel(int level)
     FAIL(0)
int snd_StartRecording(int desiredSamplesPerSec, int stereo, int semaIndex)
     FAIL(0)
int snd_StopRecording(void)
     FAIL(0)
double snd_GetRecordingSampleRate(void)
     FAIL(0)
int snd_RecordSamplesIntoAtLength(int buf, int startSliceIndex,
				  int bufferSizeInBytes)
     FAIL(0)


void snd_Volume(double *left, double *right)
{
  return;
}


void snd_SetVolume(double left, double right)
{
  return;
}


#elif defined(USE_AUDIO_OSS)


/* OSS - the "Open Sound System".  This is built into the kernel on
 * Linux, although it can be instaled on other systems as well.  See
 * <http://www.opensound.com/> for details and downloads.
 */


#ifdef HAS_MSB_FIRST
# define IS_BIGENDIAN 1
#else
# define IS_BIGENDIAN 0
#endif

#define SOUND_DEVICE		OSS_DEVICE
#define BYTES_PER_SAMPLE	2		/* Squeak always uses 16-bit samples */

#define NUM_FRAGMENTS		4	/* number of buffer fragments */
#define HWM_FRAGMENTS		2	/* play semaphore hysteresis */

#include <assert.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

static unsigned char *auBuf=	0;

static int auFd=		-1;    /* fd for /dev/dsp */
static int auReading=           0;     /* whether we are reading or writing */
static int auBytes=             2;     /* bytes per sample /dev/dsp is using */
static int auSigned=            1;     /* whether /dev/dsp uses signed samples */
static int auStereo=		0;     /* whether we are playing in stereo or not */
static int auFrameCount=	0;
static int auSampleRate=	0;
static int auSemaIndex=		0;
static int auBufBytes=		0;
static int auSwapBytes=         0;  /* whether /dev/dsp is differently-ended than Squeak */
static int auIsBigEndian=       0;  /* whether /dev/dsp is big-endion */

static int auSemaWaiting=	0;

static int mixerFd=            -1;   /* fd for /dev/mixer */



/*** Mixer routines ***/


/* close the mixer, if it is open */

static void closeMixer(void)
{
  if (mixerFd >= 0)
    {
      close(mixerFd);
      mixerFd= -1;
    }
}

/* open the mixer if it isn't already */

static void ensureMixerIsOpen(void)
{
  if (mixerFd >= 0)
    return;  /* it's already open */
  /* it's not open; try to open it now */
  mixerFd= open("/dev/mixer", O_RDWR);
}

/* query the current play volume */

void snd_Volume(double *left, double *right)
{
  int encodedVolume;

  ensureMixerIsOpen();
  /* check that the mixer is available */
  if (mixerFd < 0)
    {
      success(false);
      return;
    }
  /* read the volume */
  if (ioctl(mixerFd, SOUND_MIXER_READ_VOLUME, &encodedVolume))
    {
      success(false);
      return;
    }
  /* decode it */
  *left= (encodedVolume & 0xFF) / 100.0;
  *right= ((unsigned)(encodedVolume & 0xFF00) >> 8) / 100.0;
}

/* helper for snd_SetVolume: clamp a float between two bounds */

static double clampf(double number, double lowerBound, double upperBound)
{
  if (number < lowerBound)
    return lowerBound;
  if (number > upperBound)
    return upperBound;
  return number;
}


void snd_SetVolume(double left, double right)
{
  int encodedVolume;
  double clampedLeft, clampedRight;

  ensureMixerIsOpen();
  /* check that the mixer is available */
  if (mixerFd < 0)
    {
      success(false);
      return;
    }
  /* encode the volume */
  clampedLeft= clampf(left, 0, 1);
  clampedRight= clampf(right, 0, 1);
  encodedVolume= ((clampedRight  * 100) * 256)  +  (clampedLeft * 100);
  /* set the volume */
  if (ioctl(mixerFd, SOUND_MIXER_WRITE_VOLUME, &encodedVolume))
    {
      success(false);
      return;
    }
}


/*** /dev/dsp routines ***/


/* calculate number of bytes per frame, given the current mode */

static int bytesPerPlayFrame(void)
{
  int bytes;
     
  bytes= 1;
  bytes *= auBytes;
  if (auStereo)
    bytes *= 2;
  return bytes;
}


static void auHandle(int fd, void *data, int flags)
{
  if ( auReading && (flags & AIO_R))    signalSemaphoreWithIndex(auSemaIndex);
  if (!auReading && (flags & AIO_W))    signalSemaphoreWithIndex(auSemaIndex);
  aioHandle(fd, auHandle, flags);
}


/* reimplemented because ffs() is broken on some machines */

static int findFirstBit(int i)
{
  int p;
  for (p= 0;  i != 0;  i >>= 1) ++p;
  return p;
}


/*** exported sound output functions ***/


int snd_Stop(void)
{
  closeMixer();
  if (auFd == -1) return 0;
  aioDisable(auFd);
  close(auFd);
  auFd= -1;
  free(auBuf);
  auBuf= 0;
  auBufBytes= 0;
  return 0;
}


/* set the input/output format to use, and set the relevant au*
   variables according to what was obtained.  If no suitable format
   was set, return false, else return true. */
static int dspSetFormat(void)
{
  int formats;   /* formats available */
  int format;    /* format to use */

  /* obtain the list of available formats */
  if (ioctl(auFd, SNDCTL_DSP_GETFMTS, &formats))
    {
      perror("SNDCTL_DSP_GETFMTS");
      return false;
    }
#ifdef TRACE
  fprintf(stderr, "formats = %x\n", formats);
#endif
  /* choose a format from those available */
  if      (formats & AFMT_S16_BE) format= AFMT_S16_BE;
  else if (formats & AFMT_S16_LE) format= AFMT_S16_LE;
  else if (formats & AFMT_U16_BE) format= AFMT_U16_BE;
  else if (formats & AFMT_U16_LE) format= AFMT_U16_LE;
  else if (formats & AFMT_S8)     format= AFMT_S8;
  else if (formats & AFMT_U8)     format= AFMT_U8;
  else
    {
      fprintf(stderr, "no supported audio format (%x)\n", formats);
      return false;
    }
  /* set the desired format */
  if (ioctl(auFd, SNDCTL_DSP_SETFMT, &format))
    {
      perror("SNDCTL_DSP_SETFMT");
      return false;
    }
#ifdef TRACE  
  fprintf(stderr, "soundcard format = %d\n", format);
#endif
  /* record info on the format we ended up with */
  switch (format)
    {
    case AFMT_U16_LE:	auBytes= 2;  auIsBigEndian=  0;	 auSigned= 0;  break;
    case AFMT_U16_BE:   auBytes= 2;  auIsBigEndian=  1;	 auSigned= 0;  break;
    case AFMT_S16_LE:   auBytes= 2;  auIsBigEndian=  0;	 auSigned= 1;  break;
    case AFMT_S16_BE:   auBytes= 2;  auIsBigEndian=  1;	 auSigned= 1;  break;
    case AFMT_U8:	auBytes= 1;  auIsBigEndian= 42;	 auSigned= 0;  break;
    case AFMT_S8:	auBytes= 1;  auIsBigEndian= 42;	 auSigned= 1;  break;
    default:	/* this cannot happen */
      abort();
    }
  if (auBytes == 2)
    auSwapBytes= (auIsBigEndian != IS_BIGENDIAN);
  else
    auSwapBytes= 42;

  return true;
}


/* open the dsp device with the given parameters.  Return 0 on
   success, 1 on false.  Sets the au* variables to reflect what has
   happened. */
int dspOpen(int wantReading, int frameCount, int samplesPerSec,
	    int stereo, int semaIndex)
{
#ifdef TRACE
  fprintf(stderr, "dspOpen(wantReading=%d, frameCount=%d, samplesPerSec=%d, stereo=%d, semaIndex=%d)\n",
	  wantReading, frameCount, samplesPerSec, stereo, semaIndex);
#endif

  if (auFd != -1)
    snd_Stop();

  auReading=            wantReading;
  auStereo=		(stereo ? 1 : 0);
  auSampleRate=		samplesPerSec;
  auSemaIndex=		semaIndex;

  auFd= open(SOUND_DEVICE, O_NONBLOCK | (wantReading ? O_RDONLY : O_WRONLY), 0);
  if (auFd < 0)
    {
      perror(SOUND_DEVICE);  
      return false;
    }
  if (frameCount > 0)
    {
      int bytesPerFrame= (stereo ? 2 * BYTES_PER_SAMPLE : BYTES_PER_SAMPLE);
      int bufferBytes=	 ((frameCount * bytesPerFrame) / 8) * 8;
      int arg;
      /* round down (NOT up!!!) to nearest power of two */
      int bufferPower=	findFirstBit(bufferBytes) - 1;
      bufferBytes=	1 << bufferPower;
      auFrameCount=	bufferBytes / bytesPerFrame;
      auBufBytes=	bufferBytes;
      arg=		(NUM_FRAGMENTS << 16) | (bufferPower);

      if (ioctl(auFd, SNDCTL_DSP_SETFRAGMENT, &arg) == -1)
	{
	  perror("ioctl(SNDCTL_DSP_SETFRAGMENT)");
	  goto closeAndFail;
	}
      auBuf= (unsigned char *)malloc(bufferBytes);
    }
#if 0
  {
    struct audio_buf_info info;
    if (ioctl(auFd, SNDCTL_DSP_GETOSPACE, &info) == -1)
      {
	perror("ioctl(SNDCTL_DSP_GETOSPACE)");
	goto closeAndFail;
      }
  }
#endif
  {
    int stereoFlag= auStereo;	/* 0 => mono, 1 => stereo */

    if (ioctl(auFd, SNDCTL_DSP_STEREO, &stereoFlag) == -1)
      {
	perror("ioctl(SNDCTL_DSP_STEREO)");
	goto closeAndFail;
      }
    if (auStereo != stereoFlag)
      {
	fprintf(stderr, "snd_Start: unsupported number of channels\n");
	goto closeAndFail;
      }
  }
  if (!dspSetFormat())
    goto closeAndFail;
  {
    int speed= samplesPerSec;

    if (ioctl(auFd, SNDCTL_DSP_SPEED, &speed) == -1)
      {
	perror("ioctl(SNDCTL_DSP_SPEED)");
	goto closeAndFail;
      }  
    if (abs(speed - samplesPerSec) > (samplesPerSec / 100))
      {
	/* > 1% sample rate error */
	fprintf(stderr, "snd_Start: using %d samples/second (requested %d)\n",
		speed, samplesPerSec);
      }
    auSampleRate= speed;
  }
  /* register the sound handler with aio */
  aioEnable(auFd, 0, 0);
  aioHandle(auFd, auHandle, (auReading ? AIO_RX : AIO_WX));
           
  return true;

 closeAndFail:
  close(auFd);
  auFd= -1;
  return false;
}


int snd_Start(int frameCount, int samplesPerSec, int stereo, int semaIndex)
{
  return dspOpen(0, frameCount, samplesPerSec, stereo, semaIndex);
}


int snd_AvailableSpace(void)
{
  audio_buf_info info;
  if (auFd < 0) return -1;

  if (ioctl(auFd, SNDCTL_DSP_GETOSPACE, &info) < 0)
    {
      perror("SNDCTL_DSP_GETOSPACE");
      return 0;
    }
  /*  return (info.bytes > auBufBytes) ? auBufBytes : info.bytes;  */
  /*  return (info.fragments == 0) ? 0 : info.fragsize;  */
  if (info.fragments == 0)
    {
      auSemaWaiting= 1;
      return 0;
    }
  else
    {
      auSemaWaiting= 0;
      return info.fragsize;
    }
}


/* transform samples from Squeak's format to the sound card's format */

static int squeakToCard(unsigned char *src, int numFrames,
			unsigned char *destBuffer)
{
  unsigned char *dest;		/* next byte to write into */
  int halfFramesConsumed= 0;	/* number of half-frames used up */
  int bytesWritten= 0;		/* number of bytes written so far */

  dest= destBuffer;
     
#if defined(TRACE) || 0
  fprintf(stderr,
	  "(%d)converting %d frames from %p to %p\n",
	  utime(), numFrames, src, dest);
#endif

  while (halfFramesConsumed/2 < numFrames)
    {
      /* only write odd frames if we are in stereo */
      if (auStereo || (halfFramesConsumed%2 == 1))
	{
	  if (auBytes == 1)
	    {
	      /* 8-bit sound device; write the MSB only */
	      if (IS_BIGENDIAN)
		*dest= src[0];
	      else
		*dest= src[1];
	      /* convert it to unsigned, if necessary */
	      if (!auSigned)
		*dest= *dest + 128;
	      dest += 1;
	      bytesWritten += 1;
	    }
	  else
	    {
	      /* 16-bit sound device.  more fun */
	      unsigned char byte0= src[0], byte1= src[1];
	      /* convert to unsigned if necessary */
	      if (!auSigned)
		{
		  /* add 16384 by adding 128 to the high byte */
		  if (IS_BIGENDIAN)
		    byte0 += 128;
		  else
		    byte1 += 128;
		}
	      /* copy the bytes into the output buffer, swapping if necessary */
	      if (auSwapBytes)
		{
		  dest[0]= byte1;
		  dest[1]= byte0;
		}
	      else
		{
		  dest[0]= byte0;
		  dest[1]= byte1;
		}
	      dest += 2;
	      bytesWritten += 2;
	    }
	}
      /* even if we only skiped the sample, we have consumed
	 another half-frame and 2 bytes of source */
      halfFramesConsumed += 1;
      src += 2;
    }

#ifdef TRACE
  fprintf(stderr, "converted %d halfFrames into %d bytes\n",
	  halfFramesConsumed, bytesWritten);
#endif
  assert(halfFramesConsumed % 2 == 0);

  return halfFramesConsumed/2;
}


/* write a buffer fully out to a FD, even if it takes multiple write()
   calls.  Returns 0 on success, or -1 on failure */

static int writeFully(int fd, unsigned char *buf, int count)
{
  while (count > 0)
    {
      int len= write(auFd, buf, count);
      if (len == -1)
	{
	  if (errno != EINTR)
	    {
	      perror("write");
	      return -1;
	    }
	}
      else
	{
	  count-= len;
	  buf+= len;
	}
    }
  return 0;
}


/* play some samples out to the sound card */

int snd_PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
{
  int cardBytesAvail;
  int cardFramesAvail;
  int framesToPlay;

  cardBytesAvail= snd_AvailableSpace();
  cardFramesAvail= cardBytesAvail / bytesPerPlayFrame();

  framesToPlay= frameCount;
  if (framesToPlay > cardFramesAvail)
    framesToPlay= cardFramesAvail;

  {
    unsigned char playBuf[framesToPlay * bytesPerPlayFrame()];
    /* convert the samples */
    squeakToCard((unsigned char *)(arrayIndex + 4*startIndex),
		 framesToPlay, playBuf);
    /* write the samples out */
    if (writeFully(auFd, playBuf, sizeof(playBuf)))
      {
	success(false);
	snd_Stop();
	return 0;
      }
  }
  return framesToPlay;
}


/* This primitive is impossible to implement, since the OSS is doing
 * all the necessary buffering for us and there's no way to rewrite
 * data already written.
 *
 * (We could go the whole hog and use direct DMA access to the sound
 * drivers which would allow us to mix into a buffer already partially
 * played - but: (1) OSS only supports DMA on Linux and FreeBSD
 * derivatives, (2) direct access imposes the hardware's byte order
 * and sound format, and (3) the insertSamples call is due to vanish
 * in the near future.)
 */
int snd_InsertSamplesFromLeadTime(int frameCount, int srcBufPtr,
				  int samplesOfLeadTime)
{
  if (auFd == -1) return -1;
  /* The image says we're allowed to return 0 here, but the
     SoundPlayer barfs up a subscript bounds error.  Ho hum. */
#if 0
  return 0;	/* this is the CORRECT RESPONSE, but the image barfs */
#else
  {
    /* interim solution: play at least one buffer's worth of sound
       immediately, suspending the current sound activity.  This is
       a compromise between discarding a buffer's worth of sound
       in every cases except the very first sound to be played, and
       introducing a slight hiccup when a sound is started over the
       top of another one.  The latter is the lesser of the two
       evils. */
    int n= snd_PlaySamplesFromAtLength(frameCount, srcBufPtr, 1);
    frameCount-= n;
    if (frameCount > 0)
      n+= snd_PlaySamplesFromAtLength(frameCount, srcBufPtr, 1 + n);
    return n;
  }
#endif
}


int snd_PlaySilence(void)
{
  if (auFd == -1) return -1;
  /* nothing to do */
  return auBufBytes;
}




int snd_SetRecordLevel(int level)
{
  if (auFd < 0 || !auReading)
    {
      success(false);
      return false;
    }
  return true;
}


int snd_StartRecording(int desiredSamplesPerSec, int stereo, int semaIndex)
{
  if (!dspOpen(1, -1, desiredSamplesPerSec, stereo, semaIndex))
    return false;
  {
    int on= 0;
    if (ioctl(auFd, SNDCTL_DSP_SETTRIGGER, &on))
      {
	perror("SNDCTL_DSP_SETTRIGGER");
	return snd_Stop();
      }
    on= PCM_ENABLE_INPUT;
    if (ioctl(auFd, SNDCTL_DSP_SETTRIGGER, &on))
      {
	perror("SNDCTL_DSP_SETTRIGGER");
	return snd_Stop();
      }
  }
  return true;
}


int snd_StopRecording(void)
{
  snd_Stop();
  return true; 
}


double snd_GetRecordingSampleRate(void)
{
  if (auFd < 0 || !auReading)
    {
      success(false);
      return 0.0;
    }
  return auSampleRate;
}


/* convert from the sound card format to squeak's format */

static void cardToSqueak(unsigned char *src, int numSamples, unsigned char *dest)
{
  int samplesConsumed;
  int sample;

#ifdef TRACE
  fprintf(stderr, "convertToSqueak(%p, %d, %p)\n",
	  src, numSamples, dest);
#endif
     
  for (samplesConsumed= 0;  samplesConsumed < numSamples;  samplesConsumed++)
    {
      if (auBytes == 1)			/* 8-bit audio */
	{
	  if (auSigned)
	    sample= *src;
	  else
	    sample= *src - 128;
	  dest[0]= sample;
	  dest[1]= sample;
	  src += 1;
	}
      else				/* 16-bit */
	{
	  if (auSwapBytes)
	    {
	      dest[0]= src[1];
	      dest[1]= src[0];
	    }
	  else
	    {
	      dest[0]= src[0];
	      dest[1]= src[1];
	    }
	  if (!auSigned)
	    {
	      if (IS_BIGENDIAN)
		dest[0] -= 128;
	      else
		dest[1] -= 128;
	    }
	  src += 2;
	}
      dest += 2;
    }
}


/* return the number of bytes that are available to be read from auFd */

static int bytesAvailableToRead(void)
{
  struct audio_buf_info info;
  
  if (ioctl(auFd, SNDCTL_DSP_GETISPACE, &info) < 0)
    {
      perror("ioctl(SNDCTL_DSP_GETISPACE");
      return -1;
    }
  return info.bytes;
}



int snd_RecordSamplesIntoAtLength(int buf, int startSliceIndex,
				  int bufferSizeInBytes)
{
#ifdef TRACE
  fprintf(stderr, "recordSamples(startSliceIndex=%d, bufferSize=%d)\n",
	  startSliceIndex, bufferSizeInBytes);
#endif
  
  if (auFd < 0 || !auReading)
    return success(false);

  {
    int samplesDesired= bufferSizeInBytes / 2 - startSliceIndex;
    int cardBytesDesired= samplesDesired * auBytes;
    /* don't read more than is available */
    int cardBytesAvailable= bytesAvailableToRead();
    int cardBytesToRead;

    if (cardBytesAvailable < 0)
      return success(false);

    if (cardBytesAvailable < cardBytesDesired)
      cardBytesToRead= cardBytesAvailable;
    else
      cardBytesToRead= cardBytesDesired;
    {
      char cardBuf[cardBytesToRead];
      int cardBytes;                  /* bytes read in */
      int cardSamples;                /* number of samples that were actually read */

      /* read in the bytes */
      cardBytes= read(auFd, cardBuf, sizeof(cardBuf));
      if (cardBytes < 0)
	{
	  /* read failure of some kind */
	  if (errno == EINTR)
	    {
	      /* temporary error -- return later */
	      return 0;
	    }
	  /* serious error -- give up */
	  perror("read(auFd)");
	  return success(false);
	}
      /* copy the bytes into Squeak's buffer */
      cardSamples= cardBytes / auBytes;
      cardToSqueak(cardBuf, cardSamples, (char *) (buf + 2*startSliceIndex));
      return cardSamples;
    }
  }
}


#elif defined(USE_AUDIO_SUN)


/* SUN -- SunOS and Solaris's /dev/dsp interface.
 *
 * This support is rudimentary, and is implemented largely by reading
 * header files and guessing.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_AUDIOIO_H   /* darn, this file is in a different place between SunOS and Solaris */
#include <sys/audioio.h>
#else
#include <sun/audioio.h>
#endif
#include <errno.h>


static int auFd=		-1;  /* open on /dev/dsp */
static int auStereo=		0;   /* whether we are playing in stereo or not */
static int auPlaySemaIndex=	0;   /* an index to signal when new data may be played */
static int auBufBytes=		0;   /*  buffer size to use for playback.  unfortunately, this bears no relationship to whatever the kernal and soundcard are using  */


static int auBuffersPlayed;


static void auHandle(int fd, void *data, int flags)
{
  if (auFd < 0) return;
  if (snd_AvailableSpace() > 0)
    signalSemaphoreWithIndex(auPlaySemaIndex);
  aiohandle(fd, auHandle, flags);
}


/*** exported sound output functions ***/


int snd_Stop(void)
{
  if (auFd == -1) return;

  aioDisable(auFd);
  close(auFd);
  auFd= -1;

  return 0;
}


int snd_Start(int frameCount, int samplesPerSec, int stereo, int semaIndex)
{
  int bytesPerFrame=	(stereo ? 4 : 2);
  int bufferBytes=	((frameCount * bytesPerFrame) / 8) * 8;
  struct audio_info info;
  int err;
     
  if (auFd != -1) snd_Stop();
  auPlaySemaIndex= semaIndex;
  auStereo= stereo;
  auBufBytes= bytesPerFrame * frameCount;

  if ((auFd= open("/dev/audio", O_WRONLY)) == -1)
    {
      perror("/dev/audio");
      return false;
    }
  /* set up device */
  if (ioctl(auFd, AUDIO_GETINFO, &info))
    {
      perror("AUDIO_GETINFO");
      goto closeAndFail;
    }
  info.play.gain= 100;
  info.play.precision= 16;
  info.play.encoding= AUDIO_ENCODING_LINEAR;
  info.play.channels= auStereo ? 2 : 1;
  info.play.sample_rate= samplesPerSec;

  auBuffersPlayed= info.play.eof;

  while ((err= ioctl(auFd, AUDIO_SETINFO, &info)) && errno == EINTR)
    ;
  if (err)
    {
      perror("AUDIO_SETINFO");
      goto closeAndFail;
    }
  aioEnable(auFd, 0, 0);
  aioHandle(auFd, auHandle, AIO_WX);
  return true;
  
 closeAndFail:
  close(auFd);
  auFd= -1;
  return false;
}


int snd_AvailableSpace(void)
{
  struct audio_info info;

  if (auFd < 0) return 0;

  if (ioctl(auFd, AUDIO_GETINFO, &info))
    {
      perror("AUDIO_GETINFO");
      snd_Stop();
    }
  return (auBufBytes * (info.play.eof - auBuffersPlayed + 2));
}


int snd_PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
{
  short *src= (short *) (arrayIndex + 4*startIndex);
  short buf[2*frameCount];
  int i;
  int bytes;

  if (auFd < 0) return -1;

  if (auStereo)
    {
      bytes= 4 * frameCount;
      for (i= 0;  i < 2 * frameCount;  i++)
	buf[i]= src[i];
    }
  else
    {
      bytes= 2 * frameCount;
      for (i= 0;  i < frameCount;  i++)
	buf[i]= src[2*i];
    }
  /* write data to device from auBuf to dst */
  while (bytes > 0)
    {
      int len;
      char *pos= (char *) buf;
	  
      len= write(auFd, pos, bytes);
      if (len < 0)
	{
	  perror("/dev/audio");
	  return 0;
	}
      bytes-= len;
      pos+= len;
    }
  /* add an eof marker */
  write(auFd, buf, 0);
  auBuffersPlayed += 1;
  
  return frameCount;
}


int snd_InsertSamplesFromLeadTime(int frameCount, int srcBufPtr,
				  int samplesOfLeadTime)
{
  return 0;
}


int snd_PlaySilence(void)
{
     return 0;
}


/** recording not supported **/
int snd_SetRecordLevel(int level)
{
  success(false);
  return;
}


int snd_StartRecording(int desiredSamplesPerSec, int stereo, int semaIndex)
{
  success(false);
  return;
}


int snd_StopRecording(void)
{
  return;
}


double snd_GetRecordingSampleRate(void)
{
  success(false);
  return 0.0;
}


int snd_RecordSamplesIntoAtLength(int buf, int startSliceIndex,
				  int bufferSizeInBytes)
{
  success(false);
  return 0;
}



void snd_Volume(double *left, double *right)
{
  success(false);
  return;
}


void snd_SetVolume(double left, double right)
{
  success(false);
  return;
}


#elif defined USE_AUDIO_NONE


/* NONE -- a null device; all primitives simply fail */

#define FAIL(X) { success(false); return X; }

/* sound output */

int snd_AvailableSpace(void)
     FAIL(8192)
int snd_InsertSamplesFromLeadTime(int frameCount, int srcBufPtr,
				  int samplesOfLeadTime)
     FAIL(frameCount)
int snd_PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
     FAIL(8192)
int snd_PlaySilence(void)
     FAIL(8192)
int snd_Start(int frameCount, int samplesPerSec, int stereo, int semaIndex)
     FAIL(1)
int snd_Stop(void)
     FAIL(0)

/* sound input */

int snd_SetRecordLevel(int level)
     FAIL(0)
int snd_StartRecording(int desiredSamplesPerSec, int stereo, int semaIndex)
     FAIL(0)
int snd_StopRecording(void)
     FAIL(0)
double snd_GetRecordingSampleRate(void)
     FAIL(0)
int snd_RecordSamplesIntoAtLength(int buf, int startSliceIndex,
				  int bufferSizeInBytes)
     FAIL(0)


void snd_Volume(double *left, double *right)
{
  return;
}


void snd_SetVolume(double left, double right)
{
  return;
}


#else


# error no audio device was selected (verify your "configure" options)


#endif


/*** module initialisation/shutdown ***/


int soundInit(void)
{
  return 1;
}

int soundShutdown(void)
{
  snd_Stop();
  return 1;
}

